#include <unistd.h>
#include <stdio.h>

/* needed if you compile against linux headers */
#define mmap 	__GI_mmap
#define munmap 	__GI_munmap
#define memcpy 	__GI_memcpy
/* end */

/* heap size: for testing setted up to one page */
#define LOCAL_HEAP_SIZE 4096

/* minimun split block size */
#define MIN_FRAG_BIN	8

#define MAP_FAILED -1

/* static HEAP for testing purpose */
#ifdef ___TEST_HEAP
static char ___local_heap_bins[LOCAL_HEAP_SIZE];
#endif

/* exported Function */
int   munmap(void *addr, size_t len) __attribute__ ((visibility ("default")));
void *mmap  (void *addr, size_t len, int prot, int flags, int fd, off_t offset) __attribute__ ((visibility ("default")));
void *memcpy(void *restrict dst, const void *restrict src, size_t n) __attribute__ ((visibility ("default")));
/**/

/* memory bins struct */
typedef struct __mem_bin{
 void *prev;
 void *next;
 int addr;
 int size;
 int flags;
 int reserved;	/* don't try to remove this */
}_mem_bin;

#define BYTE_ALIGN			4
#define STAT_BLCK_FLAG_BIT 	0x00000001
#define ALLOC_BLOCK 		0x01
#define FREE_BLOCK 			0x00
#define SET_FREE_BLOCK(x) 	(x &= ~((int)0x1))
#define SET_ALLOC_BLOCK(x) 	(x |=  ((int)0x1))
#define _BLOCK_ADDR_LIM(x) 	((_mem_bin*)(x->addr + x->size))
#define SET_SIZE_ALIGN(x,y) (((x%y)+1)*y)

/* top bin elements reference */
static _mem_bin top_freed = {0, 0, 0, 0, 0};
static _mem_bin top_alloc = {0, 0, 0, 0, 0};
static _mem_bin *top_freed_blocks = &top_freed;
static _mem_bin *top_alloc_blocks = &top_alloc;

static _mem_bin *end_freed_blocks = &top_freed;
static _mem_bin *end_alloc_blocks = &top_alloc;

/* Heap manager internals */

static void __unlink_block(_mem_bin *mem_ptr)
{		
 _mem_bin *tmp_prev_blk = mem_ptr->prev;
 _mem_bin *tmp_next_blk = mem_ptr->next;
		
 tmp_prev_blk->next = tmp_next_blk;
 if (tmp_next_blk != NULL) tmp_next_blk->prev = tmp_prev_blk;
}

static void __link_block(_mem_bin *c_mem_ptr, _mem_bin *mem_ptr)
{		
 _mem_bin *tmp_next_blk = c_mem_ptr->next;
 _mem_bin *tmp_prev_blk = c_mem_ptr->prev;
 
 c_mem_ptr->next 	= mem_ptr;
 mem_ptr->prev	 	= c_mem_ptr;
 mem_ptr->next   	= tmp_next_blk;
 if (tmp_next_blk != NULL) tmp_next_blk->prev = mem_ptr;
}

static void __consolidate_block_fwd(_mem_bin *mem_ptr)
{
 _mem_bin *tmp_next_blk = mem_ptr->next;
 
 if (tmp_next_blk == NULL)
	return;
	
 if (tmp_next_blk == _BLOCK_ADDR_LIM(mem_ptr))
 {
	mem_ptr->next = tmp_next_blk->next;
	
	_mem_bin *tmp_next_next = tmp_next_blk->next;
	
	if (tmp_next_next != NULL) tmp_next_next->prev = mem_ptr;
	mem_ptr->size += (tmp_next_blk->size + sizeof(_mem_bin));
 }
}

static _mem_bin * __consolidate_block_bwd(_mem_bin *mem_ptr)
{
 _mem_bin *ret_blk = mem_ptr;
 _mem_bin *tmp_prev_blk = mem_ptr->prev;
	
 if (tmp_prev_blk == NULL || tmp_prev_blk->addr == 0)
	return ret_blk;
	
 if (mem_ptr == (_BLOCK_ADDR_LIM(tmp_prev_blk)))
 {
	tmp_prev_blk->next = mem_ptr->next;
	
	_mem_bin *tmp_next_blk = mem_ptr->next;
	
	if (tmp_next_blk != NULL) tmp_next_blk->prev = tmp_prev_blk;
	tmp_prev_blk->size += (mem_ptr->size+sizeof(_mem_bin));
	ret_blk = tmp_prev_blk;
 }
 return ret_blk;
}

static void __link_and_consolidate_block(_mem_bin *top_mem, _mem_bin *mem_ptr)
{		
 _mem_bin *curr_mem_ptr = top_mem;
 
 while(curr_mem_ptr != NULL)
 {
	_mem_bin *tmp_next_blk = curr_mem_ptr->next != NULL ? curr_mem_ptr->next : mem_ptr;
		
	if (((curr_mem_ptr->addr <= mem_ptr->addr) && (mem_ptr->addr <= tmp_next_blk->addr)) ||
		((curr_mem_ptr->addr >= mem_ptr->addr)))
	{
		/* check if we are the first fit chunk */
		if (curr_mem_ptr->addr >= mem_ptr->addr)
		{
			_mem_bin *tmp = curr_mem_ptr;
			curr_mem_ptr = mem_ptr;
			mem_ptr = tmp;
		}
		__link_block(curr_mem_ptr, mem_ptr);
		mem_ptr = __consolidate_block_bwd(mem_ptr);
		__consolidate_block_fwd(mem_ptr);
		break;
	}
	curr_mem_ptr = curr_mem_ptr->next;
 }
}

static void __split_bin_blk(_mem_bin *c_mem_ptr, size_t len)
{
 /* try to split the bin: min size + sizeof overhead */
 if (c_mem_ptr->size >= (sizeof(_mem_bin) + MIN_FRAG_BIN))
 {	
	_mem_bin *tmp_new_blk  = (_mem_bin*)((char*)c_mem_ptr->addr + len);

	tmp_new_blk->flags = 0;
	tmp_new_blk->size  = c_mem_ptr->size - (len + sizeof(_mem_bin));
	tmp_new_blk->addr  = (int)tmp_new_blk + sizeof(_mem_bin);
	tmp_new_blk->prev  = (void*)c_mem_ptr;
	tmp_new_blk->next  = c_mem_ptr->next;
	
	c_mem_ptr->size = len;
	c_mem_ptr->next = tmp_new_blk; 
 }
}

static void __init_heap_blks(void)
{
 if (top_freed_blocks->next == NULL)
 {
	/* static heap for testing purpose */
#ifdef ___TEST_HEAP
	top_freed_blocks->next  = ___local_heap_bins;
	_mem_bin *tmp_top_freed_blk = (_mem_bin*)___local_heap_bins;
	/* */
#else
	/* setup your heap zone */
#endif

#ifdef ___TEST_HEAP
	tmp_top_freed_blk->addr  = (int)___local_heap_bins + sizeof(_mem_bin);
#else
	/* setup your first block bin addr */
#endif

	tmp_top_freed_blk->size  = LOCAL_HEAP_SIZE;
	tmp_top_freed_blk->flags = 0;
	tmp_top_freed_blk->prev  = top_freed_blocks;
	tmp_top_freed_blk->next  = NULL;
 }
}

/*
 * Heap mapping functions
 */
 
int munmap(void *addr, size_t len)
{
 _mem_bin *mem_ptr = top_alloc_blocks->next;
 
 while (mem_ptr != NULL)
 {
	/* compat issue don't try to remove sizeof(size_t) */
	if( ((addr + sizeof(size_t)) == (void*)mem_ptr->addr) && 
	    (mem_ptr->flags & STAT_BLCK_FLAG_BIT) == ALLOC_BLOCK)
	{
		SET_FREE_BLOCK(mem_ptr->flags);
		__unlink_block(mem_ptr);
		__link_and_consolidate_block(top_freed_blocks->next, mem_ptr);
		return 0;
	}
	mem_ptr = mem_ptr->next;
 }
 return -1;
}

void *mmap(void *addr, size_t len, int prot, int flags, int fd, off_t offset)
{
 void *ret_addr = (void*)MAP_FAILED;
 int  found = 0;
 
 SET_SIZE_ALIGN(len, BYTE_ALIGN);
 
 if (len == 0)
	return ret_addr;

 if (top_freed_blocks->next == NULL)
	__init_heap_blks();
 	
 _mem_bin *mem_ptr   = top_freed_blocks->next;
 _mem_bin *c_mem_ptr = top_freed_blocks->next;
 
 while (mem_ptr != NULL)
 {
	if (len <= mem_ptr->size && (mem_ptr->size <= c_mem_ptr->size))
	{
		c_mem_ptr = mem_ptr; 
		found = 1;
	}
	mem_ptr = mem_ptr->next;
 }
 
 /* checking if we are in list element or we are not able to alloc anymore */
 if (found == 0)
	return ret_addr;
	
 __split_bin_blk(c_mem_ptr, len);
 __unlink_block(c_mem_ptr);
 __link_block(end_alloc_blocks, c_mem_ptr);
 
 end_alloc_blocks = c_mem_ptr;

 SET_ALLOC_BLOCK(c_mem_ptr->flags);
 ret_addr = (void*)c_mem_ptr->addr;
 
 /* compat issue don't try to remove sizeof(size_t) */
 return ret_addr - sizeof(size_t);
}

void *memcpy(void *restrict dst, const void *restrict src, size_t n)
{
 char *tmpdst = (char*)dst;
 char *tmpsrc = (char*)src;
 
 for (int i=0; i < n; i++)
	tmpdst[i] = tmpsrc[i];
	
 return dst;
}